from ast import Call
from dataclasses import dataclass
from typing import Callable

from django.db.models import Model
from django.urls import URLPattern, path

"""
ACTIONS
Used in models.py and in the urls.py by the convenience dataclass declared below.
"""
ADD_TAGS = "add_tags"
DEL_TAG = "del_tag"
TOGGLE_STATUS = "toggle_status"
LAUNCH_MODAL = "launch_modal"
GET_ITEM = "get_item"


"""
TEMPLATES
Placed here for easier access in configuration and tests)
"""

MODAL_BASE = "commons/modal.html"
PANEL = "commons/_panel.html"
ITEM = "commons/item.html"
LIST_BOOKMARKED = "bookmarks/bookmark_objs.html"
LIST_TAGS = "tags/list_of_tags.html"
LIST_FILTERED = "tags/filtered_objs.html"

"""
URLS
"""


@dataclass
class Pathmaker:
    """Auto-generate patterns with the convention `{act}_{model_klass._meta.model_name}` from view functions with respect to: (a) launching a modal inspecting a specific `obj` instance of a `model_klass`  (b) adding / deleting tags within the panel with respect to such `obj` instance; and (c) bookmarking / unbookmarking the `obj`."""

    model_klass: Model
    launch_func: Callable
    get_item_func: Callable
    add_tags_func: Callable
    del_tag_func: Callable
    toggle_status_func: Callable

    def make_patterns(self) -> list[URLPattern]:
        return [  # fake implies we don't know the value of the arg yet
            self.uniform_act_path(LAUNCH_MODAL, self.launch_func, fake=True),
            self.uniform_act_path(LAUNCH_MODAL, self.launch_func),
            self.uniform_act_path(GET_ITEM, self.get_item_func, fake=True),
            self.uniform_act_path(GET_ITEM, self.get_item_func),
            self.uniform_act_path(ADD_TAGS, self.add_tags_func),
            self.uniform_act_path(DEL_TAG, self.del_tag_func),
            self.uniform_act_path(TOGGLE_STATUS, self.toggle_status_func),
        ]

    def uniform_act_path(
        self, act: str, func: Callable, fake: bool = False
    ) -> URLPattern:
        """
        Given parameters for launching the modal with a view, produces the following path:
        - `route` = samplebook/launch_modal
        - `view` = launch_func (the function object Callable, not its return)
        - `name` = launch_modal_samplebook (the combiation of act and model name)

        Note nuance addition: fake urls imply the need to make the url dynamic such as when a modal url will only be filled up with an identifier during runtime or when it's necessary to setup the base url before adding from the ORM.
        """
        model_name = self.model_klass._meta.model_name
        route = f"{model_name}/{act}/"
        if not fake:
            route += "<str:pk>"
        return path(route=route, view=func, name=f"{act}_{model_name}")



# Generated by Django 4.0.1 on 2022-02-04 08:07

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('examples', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='samplequote',
            name='book',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='quotes', to='examples.samplebook'),
        ),
    ]



from django.urls import path

from .views import BOOK, QUOTE, SampleBookDetail, homepage_view

app_name = "examples"
urlpatterns = (
    BOOK.make_patterns()
    + QUOTE.make_patterns()
    + [
        path(
            "book/detail/<int:pk>",
            SampleBookDetail.as_view(),
            name="book_detail",
        ),
        path("", homepage_view),
    ]
)



# Generated by Django 4.0.1 on 2022-01-31 05:27

import django.db.models.deletion
import django_extensions.db.fields
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="TagItem",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created",
                    django_extensions.db.fields.CreationDateTimeField(
                        auto_now_add=True, verbose_name="created"
                    ),
                ),
                (
                    "modified",
                    django_extensions.db.fields.ModificationDateTimeField(
                        auto_now=True, verbose_name="modified"
                    ),
                ),
                ("name", models.SlugField(max_length=100)),
            ],
            options={
                "ordering": ["-created"],
            },
        ),
        migrations.CreateModel(
            name="Bookmark",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created",
                    django_extensions.db.fields.CreationDateTimeField(
                        auto_now_add=True, verbose_name="created"
                    ),
                ),
                (
                    "modified",
                    django_extensions.db.fields.ModificationDateTimeField(
                        auto_now=True, verbose_name="modified"
                    ),
                ),
                ("object_id", models.CharField(max_length=250)),
                (
                    "bookmarker",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "tags",
                    models.ManyToManyField(
                        related_name="bookmarked", to="bookmarks.TagItem"
                    ),
                ),
            ],
            options={
                "get_latest_by": "modified",
                "abstract": False,
            },
        ),
    ]



from django.contrib import admin

from .models import SampleBook, SampleQuote


class SampleBookAdmin(admin.ModelAdmin):
    fields = ["title", "excerpt", "author"]


admin.site.register(SampleBook, SampleBookAdmin)


class SampleQuoteAdmin(admin.ModelAdmin):
    fields = ["book", "quote"]


admin.site.register(SampleQuote, SampleQuoteAdmin)



from django.contrib.auth.decorators import login_required
from django.contrib.contenttypes.models import ContentType
from django.http import HttpRequest, HttpResponse
from django.shortcuts import get_object_or_404
from django.template.response import TemplateResponse
from django.views.decorators.http import (
    require_GET,
    require_http_methods,
    require_POST,
)
from django.views.generic import DetailView

from bookmarks.utils import ITEM, MODAL_BASE, PANEL, Pathmaker

from .models import SampleBook, SampleQuote


def homepage_view(request: HttpRequest):
    context = {
        "book_list": SampleBook.objects.all(),
        "quote_list": SampleQuote.objects.all(),
    }
    return TemplateResponse(request, "home.html", context)


class SampleBookDetail(DetailView):
    model = SampleBook
    template_name = "book_detail.html"
    context_object_name = "book"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["quotes_saved"] = self.get_object().quotes.filter(
            id__in=list(
                self.request.user.bookmark_set.filter(
                    content_type=ContentType.objects.get(
                        app_label="examples", model="samplequote"
                    )
                ).values_list("object_id", flat=True)
            )
        )
        return context


"""
? BOOKS
"""


@login_required
@require_GET
def launch_modal_samplebook(request: HttpRequest, pk: str) -> TemplateResponse:
    obj = get_object_or_404(SampleBook, pk=pk)
    panel = {"content_template": PANEL}
    context = obj.set_bookmarked_context(request.user) | panel
    return TemplateResponse(request, MODAL_BASE, context)


@login_required
@require_GET
def get_item_samplebook(request: HttpRequest, pk: str) -> TemplateResponse:
    obj = get_object_or_404(SampleBook, pk=pk)
    context = obj.set_bookmarked_context(request.user)
    return TemplateResponse(request, PANEL, context)


@login_required
@require_http_methods(["PUT"])
def toggle_status_samplebook(
    request: HttpRequest, pk: str
) -> TemplateResponse:
    obj = get_object_or_404(SampleBook, pk=pk)
    obj.toggle_bookmark(request.user)
    context = obj.set_bookmarked_context(request.user)
    return TemplateResponse(request, PANEL, context)


@login_required
@require_POST
def add_tags_samplebook(request: HttpRequest, pk: str) -> TemplateResponse:
    obj = get_object_or_404(SampleBook, pk=pk)
    if submitted := request.POST.get("tags"):
        if add_these := submitted.split(","):
            obj.add_tags(request.user, add_these)
    context = obj.set_bookmarked_context(request.user)
    return TemplateResponse(request, PANEL, context)


@login_required
@require_http_methods(["DELETE"])
def del_tag_samplebook(request: HttpRequest, pk: str) -> HttpResponse:
    obj = get_object_or_404(SampleBook, pk=pk)
    if delete_this := request.POST.get("tag"):
        obj.remove_tag(request.user, delete_this)
    return HttpResponse(headers={"HX-Trigger": "tagDeleted"})


BOOK = Pathmaker(
    model_klass=SampleBook,
    launch_func=launch_modal_samplebook,
    get_item_func=get_item_samplebook,
    toggle_status_func=toggle_status_samplebook,
    add_tags_func=add_tags_samplebook,
    del_tag_func=del_tag_samplebook,
)

"""
? QUOTES
"""


@login_required
@require_GET
def launch_modal_samplequote(
    request: HttpRequest, pk: str
) -> TemplateResponse:
    obj = get_object_or_404(SampleQuote, pk=pk)
    panel = {"content_template": PANEL}
    context = obj.set_bookmarked_context(request.user) | panel
    return TemplateResponse(request, MODAL_BASE, context)


@login_required
@require_GET
def get_item_samplequote(request: HttpRequest, pk: str) -> TemplateResponse:
    obj = get_object_or_404(SampleQuote, pk=pk)
    context = obj.set_bookmarked_context(request.user)
    return TemplateResponse(request, PANEL, context)


@login_required
@require_http_methods(["PUT"])
def toggle_status_samplequote(
    request: HttpRequest, pk: str
) -> TemplateResponse:
    obj = get_object_or_404(SampleQuote, pk=pk)
    obj.toggle_bookmark(request.user)
    context = obj.set_bookmarked_context(request.user)
    return TemplateResponse(request, PANEL, context)


@login_required
@require_POST
def add_tags_samplequote(request: HttpRequest, pk: str) -> TemplateResponse:
    obj = get_object_or_404(SampleQuote, pk=pk)
    if submitted := request.POST.get("tags"):
        if add_these := submitted.split(","):
            obj.add_tags(request.user, add_these)
    context = obj.set_bookmarked_context(request.user)
    return TemplateResponse(request, PANEL, context)


@login_required
@require_http_methods(["DELETE"])
def del_tag_samplequote(request: HttpRequest, pk: str) -> HttpResponse:
    obj = get_object_or_404(SampleQuote, pk=pk)
    if delete_this := request.POST.get("tag"):
        obj.remove_tag(request.user, delete_this)
    return HttpResponse(headers={"HX-Trigger": "tagDeleted"})


QUOTE = Pathmaker(
    model_klass=SampleQuote,
    launch_func=launch_modal_samplequote,
    get_item_func=get_item_samplequote,
    toggle_status_func=toggle_status_samplequote,
    add_tags_func=add_tags_samplequote,
    del_tag_func=del_tag_samplequote,
)



from typing import Optional

from django.contrib.auth import get_user_model
from django.contrib.contenttypes.fields import (
    GenericForeignKey,
    GenericRelation,
)
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models.query import QuerySet
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.html import format_html
from django.utils.safestring import SafeText
from django.utils.text import slugify
from django_extensions.db.models import TimeStampedModel

from .managers import MarkedTags, UserAnnotations
from .utils import ADD_TAGS, DEL_TAG, GET_ITEM, LAUNCH_MODAL, TOGGLE_STATUS


class TagItem(TimeStampedModel):
    name = models.SlugField(max_length=100)

    # managers
    objects = models.Manager()
    tagged = UserAnnotations()

    def __str__(self) -> str:
        return self.name

    class Meta:
        ordering = ["-created"]

    @classmethod
    def set_context(cls, user, tag_slug: str, model_id: Optional[int] = None):
        """The `tag_slug` (and optional `model_id`) from url refer to objects requesting `user` has previously tagged. If user is not authenticated, show empty list."""
        context = {}
        if model_id:
            context["model_type"] = ContentType.objects.get_for_id(model_id)
        tag = get_object_or_404(TagItem, name=tag_slug)
        extract = Bookmark.objects_tagged.extract_from
        context["user_tagged_objs"] = extract(user, tag, model_id)
        return context


class Bookmark(TimeStampedModel):
    # main fields
    bookmarker = models.ForeignKey(get_user_model(), on_delete=models.PROTECT)
    tags = models.ManyToManyField(TagItem, related_name="bookmarked")

    # generic fk base
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.CharField(max_length=250)  # allows UUID
    content_object = GenericForeignKey("content_type", "object_id")

    # managers
    objects = models.Manager()
    objects_tagged = MarkedTags()

    def __str__(self):
        return f"{self.bookmarker} saved {self.content_object}"


class AbstractBookmarkable(models.Model):
    bookmarks = GenericRelation(
        Bookmark, related_query_name="%(app_label)s_%(class)ss"
    )

    class Meta:
        abstract = True

    def make_action_url(self, act: str):
        base = f"{self._meta.app_label}:{act}_{self._meta.model_name}"
        return reverse(base, args=(self.pk,))

    @cached_property
    def launch_modal_url(self):
        return self.make_action_url(LAUNCH_MODAL)

    @cached_property
    def get_item_url(self):
        return self.make_action_url(GET_ITEM)

    @cached_property
    def add_tags_url(self):
        return self.make_action_url(ADD_TAGS)

    @cached_property
    def del_tag_url(self):
        return self.make_action_url(DEL_TAG)

    @cached_property
    def toggle_status_url(self):
        return self.make_action_url(TOGGLE_STATUS)

    @property
    def modal(self) -> SafeText:
        """Return html with htmx modal launcher based on app_name. Presumes prior coordination with bookmarks.utils in urls.py"""
        raw = """<em hx-trigger="click"
                hx-get="{url}"
                hx-target="body"
                hx-swap="beforeend"
                hx-indicator="#{loader_id}"
                _="on mouseover add [@style=text-decoration:underline] to me
                    on mouseleave remove [@style=text-decoration:underline] from me
                "
                >view</em>
                <span id="{loader_id}" class="spinner-border spinner-border-sm htmx-indicator" role="status">
                    <span class="visually-hidden">Loading...</span>
                </span>
            """
        return format_html(
            raw, loader_id=f"spinner-{self.pk}", url=self.launch_modal_url
        )

    @property
    def object_content_for_panel(self) -> SafeText:
        """Displayed at the top of the tag PANEL."""
        raw = "<h2>Must be overriden by {{model_name}}'s properties.</h2>"
        return format_html(raw, model_name=self._meta.model_name)

    def set_bookmarked_context(self, user) -> dict:
        """The tag PANEL in bookmarks/utils.py requires the use of certain variables that will not change, e.g. `is_bookmarked`, `toggle_url`. The values that fill these constants however will change based on the object instance `obj` and the `user` that is passed to this method."""
        return {
            "object": self,
            "object_content_for_panel": self.object_content_for_panel,
            "is_bookmarked": self.is_bookmarked(user),
            "user_tags": self.get_user_tags(user),
            "toggle_url": self.toggle_status_url,
            "add_tags_url": self.add_tags_url,
            "del_tag_url": self.del_tag_url,
        }

    def is_bookmarked(self, user) -> bool:
        """Has `user` bookmarked to this object instance?"""
        return self.bookmarks.filter(bookmarker=user).exists()

    def get_bookmarked(self, user) -> Optional[Bookmark]:
        """Get instance bookmarked to by the `user`."""
        if not self.is_bookmarked(user):
            return None
        return self.bookmarks.get(bookmarker=user)

    def get_user_tags(self, user) -> QuerySet:
        """If user bookmarked, get `user`-made tags on instance."""
        if bookmark := self.get_bookmarked(user):
            return bookmark.tags.all()
        return []

    def toggle_bookmark(self, user) -> bool:
        """If `user` is bookmarked to the instance, unbookmark; otherwise, bookmark."""
        if not self.is_bookmarked(user):
            return self._bookmark_this(user)
        return self._unbookmark_this(user)

    def _unbookmark_this(self, user) -> bool:
        """Implies `user` already bookmarked to the instance. This removes the `bookmark` object from the instance's `bookmarks` field."""
        bookmark = self.get_bookmarked(user)
        self.bookmarks.remove(bookmark)
        if tags := self.get_user_tags(user):
            tags.delete()
        return self.is_bookmarked(user)  # status after unbookmarking

    def _bookmark_this(self, user) -> bool:
        """Implies `user` is not yet bookmarked to the instance. This add a `bookmark` object to the instance's `bookmarks` field."""
        bookmark = Bookmark(content_object=self, bookmarker=user)
        self.bookmarks.add(bookmark, bulk=False)
        return self.is_bookmarked(user)  # status after bookmark

    def add_tags(self, user, tags_to_add: list[str]):
        """Parse a list of `tags_to_add`, by a `user` to an auto-bookmarked model instance."""
        if not self.is_bookmarked(user):  # auto-bookmark
            self._bookmark_this(user)

        bookmark = self.bookmarks.get(bookmarker=user)
        _existing = list(bookmark.tags.values_list("name", flat=True))
        for input_name in tags_to_add:
            slug = slugify(input_name)
            if slug not in _existing:
                tag_obj, _ = TagItem.objects.get_or_create(name=slug)
                bookmark.tags.add(tag_obj)

    def remove_tag(self, user, tag_to_remove: str):
        """Since bookmarked instance can have existing tags, enable user to remove an existing tag name."""
        slug = slugify(tag_to_remove)
        tag_to_remove = get_object_or_404(TagItem, name=slug)

        if not self.is_bookmarked(user):  # auto-bookmark
            self._bookmark_this(user)

        bookmark = self.bookmarks.get(bookmarker=user)
        if bookmark.tags.filter(name=slug).exists():
            bookmark.tags.remove(tag_to_remove)



from django.urls import path

from .views import annotated_tags, bookmarked_objs, filtered_objs

app_name = "bookmarks"
urlpatterns = [
    path(
        "tag/<slug:tag_slug>/<int:model_id>",
        filtered_objs,
        name="filtered_objs",
    ),
    path("tag/<slug:tag_slug>", filtered_objs, name="filtered_objs"),
    path("tags", annotated_tags, name="annotated_tags"),
    path("objs", bookmarked_objs, name="bookmarked_objs"),
]



# Generated by Django 4.0.1 on 2022-01-31 05:27

import uuid

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="SampleBook",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=50)),
                ("excerpt", models.TextField(null=True)),
                (
                    "author",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "Book",
                "verbose_name_plural": "Books",
            },
        ),
        migrations.CreateModel(
            name="SampleQuote",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                ("quote", models.TextField()),
                (
                    "book",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="examples.samplebook",
                    ),
                ),
            ],
            options={
                "verbose_name": "Quote",
                "verbose_name_plural": "Quotes",
            },
        ),
    ]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bookmarks', 'bookmarks.migrations', 'examples', 'examples.migrations']

package_data = \
{'': ['*'],
 'bookmarks': ['docs/*',
               'static/css/*',
               'static/img/favicons/*',
               'templates/*',
               'templates/bookmarks/*',
               'templates/commons/*',
               'templates/tags/*'],
 'examples': ['fixtures/*', 'templates/*']}

install_requires = \
['Django>=4.0,<5.0',
 'django-debug-toolbar>=3.2.2,<4.0.0',
 'django-extensions>=3.1.5,<4.0.0']

setup_kwargs = {
    'name': 'django-yabl',
    'version': '0.0.6',
    'description': 'Yet another bookmarking library (yabl) for Django. Bookmark and tag arbitrary models.',
    'long_description': "# django-yabl\n\nYet another bookmarking library (yabl) for Django. Bookmark and tag arbitrary models.\n\n`AbstractBookmarkable` contains `bookmarks` field. This enables arbitrary child models, e.g. Movies, Books, Laws, Clothes, etc., to inherit uniform properties for bookmarking and tagging.\n\nThe `bookmarks` field is mapped to a generic `Bookmark` model containing:\n\n1. the authenticated user adding the bookmark, i.e. the `bookmarker`;\n2. the concrete model instance referenced, i.e. the _bookmarked_;[^1] and\n3. a ManyToMany `tags` field which maps to a `TagItem` model.\n\n[^1]: The model is referenced via a `content_type` and an `object_id`\n\n## AbstractBookmarkable\n\nThe abstraction makes each inheriting instance _bookmarkable_ and _taggable_ by authenticated users.\n\n| Attributes                        | Purpose                                                    |\n| --------------------------------- | ---------------------------------------------------------- |\n| `is_bookmarked`(user)             | Check whether object instance is bookmarked or not         |\n| `get_bookmarked`(user)            | Get instances of model that user has bookmarked            |\n| `get_user_tags`(user)             | If user bookmarked, get user-made tags on instance         |\n| `toggle_bookmark`(user)           | Toggle bookmark status as bookmarked or not                |\n| `add_tags`(user, tags: list[str]) | Add unique tags, accepts a list of names                   |\n| `remove_tag`(user, tag: str)      | Delete an existing tag name from tags previously added     |\n| `set_bookmarked_context`(user)    | Combines relevant urls and attributes for template output  |\n| @`modal`                          | Custom modal enables: _toggle_, _add tags_, _remove tag_   |\n| @`launch_modal_url`               | URL to launch custom modal                                 |\n| @`get_item_url`               | URL to load the panel containde within custom modal                                 |\n| @`add_tags_url`                   | URL to POST tags added                                     |\n| @`del_tag_url`                    | URL to DELETE tag added                                    |\n| @`toggle_status_url`              | URL to toggle bookmark status of an object instance added  |\n| @`object_content_for_panel`       | Content when custom modal is loaded; **must** be overriden |\n\n## Modal-based UX\n\n### What is the concept?\n\nThe modal is where the user interacts – i.e. toggle bookmark status, add tags, remove tag (see table above) – with the data.\n\n### How is the modal styled?\n\nSee the htmx/hyperscript example [modal.css](bookmarks/static/css/modal.css).\n\n### How is the modal constructed during run-time?\n\nSee _app-level_ [modal.html](bookmarks/templates/commons/modal.html) which shows a modal via htmx click on the `@launch_modal_url` property.\n\n### What are the pre-made contents of the modal?\n\nThe _app-level_ [panel.html](bookmarks/templates/tags/templates/tags/panel.html), contained within the modal, shows an actionable form for saving the bookmarable object and associating said object with tags:\n\n1. The submission of tags is POST'ed through the `add_tags_url`.\n2. The deletion of tags is DELETE'ed through the `del_tag_url`.\n3. The bookmark toggle is PUT'ed through the `toggle_status_url`.\n\n## Setup\n\n1. Download and [install](bookmarks/docs/setup.md).\n2. See [configuration](bookmarks/docs/configure.md) of models to be bookmarked and tagged.\n3. Examine [frontend](bookmarks/docs/frontend.md) setup using `htmx/hyperscript`.\n",
    'author': 'Marcelino G. Veloso III',
    'author_email': 'mars@veloso.one',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/justmars/django-yabl',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import Optional

from django.http import HttpRequest
from django.template.response import TemplateResponse

from .models import TagItem
from .utils import LIST_BOOKMARKED, LIST_FILTERED, LIST_TAGS


def filtered_objs(
    request: HttpRequest, tag_slug: str, model_id: Optional[int] = None
) -> TemplateResponse:
    """Get objects tagged with `tag_slug`, optionally filtered by `model_id`, assuming user is authenticated."""
    context = {"user_tagged_objs": [], "tag_slug": tag_slug}
    if request.user.is_authenticated:
        context |= TagItem.set_context(request.user, tag_slug, model_id)
    return TemplateResponse(request, LIST_FILTERED, context)


def annotated_tags(request: HttpRequest) -> TemplateResponse:
    from .models import AbstractBookmarkable

    tags = []
    if request.user.is_authenticated:
        tags = TagItem.tagged.made_by_user(
            request.user, AbstractBookmarkable.__subclasses__()
        )
    return TemplateResponse(request, LIST_TAGS, {"tags": tags})


def bookmarked_objs(request: HttpRequest) -> TemplateResponse:
    objs = []
    if request.user.is_authenticated:
        objs = request.user.bookmark_set.all()
    context = {"bookmarked_objs": objs}
    return TemplateResponse(request, LIST_BOOKMARKED, context)



import uuid

from django.contrib.auth import get_user_model
from django.db import models
from django.urls import reverse
from django.utils.html import format_html
from django.utils.safestring import SafeText

from bookmarks.models import AbstractBookmarkable


class SampleBook(AbstractBookmarkable):
    """Used for testing only. See bookmarks/tests/conftest.py"""

    title = models.CharField(max_length=50)
    excerpt = models.TextField(null=True)
    author = models.ForeignKey(get_user_model(), on_delete=models.PROTECT)

    class Meta:
        verbose_name = "Book"  # see generic relations, e.g. content_type.name
        verbose_name_plural = "Books"

    def __str__(self) -> str:
        return (
            f"{self.title} by {self.author.first_name} {self.author.last_name}"
        )

    def get_absolute_url(self):
        return reverse("examples:book_detail", kwargs={"pk": self.pk})

    @property
    def object_content_for_panel(self) -> SafeText:
        return format_html(
            """
            <h2>{title}</h2>
            <h3>{author}</h3>
            <p>{excerpt}</p>
            """,
            title=self.title,
            excerpt=self.excerpt,
            author=f"{self.author.first_name} {self.author.last_name}",
        )


class SampleQuote(AbstractBookmarkable):
    """Used for testing only. See bookmarks/tests/conftest.py"""

    id = models.UUIDField(primary_key=True, default=uuid.uuid4)
    book = models.ForeignKey(
        SampleBook, on_delete=models.CASCADE, related_name="quotes"
    )
    quote = models.TextField()

    class Meta:
        verbose_name = "Quote"  # see generic relations, e.g. content_type.name
        verbose_name_plural = "Quotes"

    def __str__(self) -> str:
        return f"{self.quote[:50]}"

    @property
    def object_content_for_panel(self) -> SafeText:
        return format_html(
            """
            <figure>
                <blockquote>{quote}</blockquote>
                <center>
                    <figcaption>
                    {author}, <cite title="{book_title}">{book_title}</cite>
                    </figcaption>
                </center>
            </figure>
            """,
            book_title=self.book.title,
            quote=self.quote,
            author=f"{self.book.author.first_name} {self.book.author.last_name}",
        )



from typing import Optional

from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import Count, Q, Value
from django.db.models.query import QuerySet


class UserAnnotations(models.Manager):
    def filter_by_user(self, user) -> QuerySet:
        """Get all tags in which the `user` has bookmarked to a bookmarked model instance."""
        qs = (
            super()
            .get_queryset()
            .prefetch_related("bookmarked__bookmarker")
            .prefetch_related("bookmarked__content_type")
            .filter(bookmarked__bookmarker=user)
            .distinct()
        )
        return qs

    def prep_for_annotation(self, user, model: models.Model):
        """Using the prefix of the model `label`, generate values for `_count`, `_id`, and `_slug`.

        e.g if the model is `SampleBook`, the generated keys are `samplebook_count`, `samplebook_id`, and `samplebook_slug`.

        1. The `_id` refers to the `contenttype` id.
        2. The `_slug` refers to the lowercased name of the model.
        3. The `_count` refers to the number of instances bookmarked per type.

        See tags/tagged_models.html for how used."""
        label = model._meta.model_name.lower()
        type_of_model = ContentType.objects.get_for_model(model)
        by_type = Q(bookmarked__content_type=type_of_model)
        by_user = Q(bookmarked__bookmarker=user)
        return {
            f"{label}_count": Count("bookmarked", filter=by_type & by_user),
            f"{label}_id": Value(type_of_model.id),
            f"{label}_slug": Value(model._meta.verbose_name),
        }

    def made_by_user(self, user, models: list[models.Model]) -> QuerySet:
        """Tags filtered by user then each tag is annotated with different contenttypes."""
        qs = self.filter_by_user(user)
        for model in models:
            values = self.prep_for_annotation(user, model)
            qs = qs.annotate(**values)
        return qs


class MarkedTags(models.Manager):
    def _bookmarker_by_user(self, user):
        """Each user may have bookmarked objects. This fetches all bookmarks made by a specific user."""
        return (
            super()
            .get_queryset()
            .select_related("bookmarker")
            .prefetch_related("tags")
            .filter(bookmarker=user)
            .distinct()
        )

    def extract_from(
        self, user, tag, content_id: Optional[int] = None
    ) -> QuerySet:
        """bookmarks tagged by the `user` with a `tag_slug`. The resulting queryset can be of different models since the `bookmark` model is generic; but if `content_id` is set, the instances must be of the same contenttype model represented by the `content_id`."""
        qs = self._bookmarker_by_user(user).filter(tags=tag)
        if content_id:
            qs = qs.filter(
                content_type=ContentType.objects.get_for_id(content_id)
            )
        return qs
